/*-------------- Telecommunications & Signal Processing Lab ---------------

Routine:
  AFILE *AFrdAIhead (FILE *fp)

Purpose:
  Get file format information from an AIFF or AIFF-C audio file

Description:
  This routine reads the header for an AIFF or AIFF-C audio file.  The header
  information is used to set the file data format information in the audio
  file pointer structure.

  AIFF or AIFF-C audio file:
   Offset Length Type    Contents
      0     4    char   File identifier ("FORM")
      4     4    int    Form chunk length
      8     4    char   File identifier ("AIFF" or "AIFC")
    ...   ...    ...    ...
      C     4    char   COMM chunk identifier ("COMM")
    C+4     4    int    COMM chunk length
    C+8     2    int    Number of interleaved channels
    C+10    4    int    Number of sample frames
    C+14    2    int    Bits per sample
    C+16   10    float  Sample frames per second
    C+26    4    int    Compression type
    C+30  ...    char   Compression name
    ...    ...   ...    ...
      S     4    char   SSND chunk identifier ("SSND")
    S+4     4    int    SSND chunk length
    S+8     4    int    Data offset
    S+12    4    int    Block size
    S+16  ...    ...    Audio data

Parameters:
  <-  AFILE *AFrdAIhead
      Audio file pointer for the audio file
   -> FILE *fp
      File pointer for the file

Author / revision:
  P. Kabal  Copyright (C) 1999
  $Revision: 1.52 $  $Date: 1999/06/04 20:07:54 $

-------------------------------------------------------------------------*/

static char rcsid [] = "$Id: AFrdAIhead.c 1.52 1999/06/04 AFsp-v4r3 $";

#include <setjmp.h>
#include <string.h>

#include <libtsp.h>
#include <libtsp/nucleus.h>
#include <libtsp/AFdataio.h>
#include <libtsp/AFheader.h>
#include <libtsp/AFmsg.h>
#include <libtsp/AFpar.h>
#include <libtsp/AIpar.h>

#define ICEILV(n,m)	(((n) + ((m) - 1)) / (m))	/* int n,m >= 0 */
#define MINV(a, b)	(((a) < (b)) ? (a) : (b))
#define RNDUPV(n,m)	((m) * ICEILV (n, m))		/* Round up */

#define SAME_CSTR(str,ref) 	(memcmp (str, ref, sizeof (str)) == 0)

#define ALIGN		2	/* Chunks padded out to a multiple of ALIGN */

/* setjmp / longjmp environment */
extern jmp_buf AFR_JMPENV;

static long int
AF_rdFORM_AIFF (FILE *fp, int *Ftype);
static int
AF_rdCOMM (FILE *fp, int Size, int Ftype, struct AI_CkCOMM *CkCOMM);
static void
AF_decCOMM (struct AI_CkCOMM *CkCOMM, struct AF_read *AFr);
static int
AF_checkFVER (FILE *fp, int Size);
static long int
AF_decSSND (FILE *fp, long int Size, long int *Ldata);
static int
AF_rPstring (FILE *fp, char string[], int ncMax);


AFILE *
AFrdAIhead (FILE *fp)

{
  AFILE *AFp;
  long int offs, LFORM, Dstart;
  int Ftype, AtData, N;
  char Info[AF_MAXINFO];
  struct AF_read AFr;

  struct AI_CkPreamb CkHead;
  struct AI_CkCOMM CkCOMM;

/* Set the long jump environment; on error return a NULL */
  if (setjmp (AFR_JMPENV))
    return NULL;	/* Return from a header read error */

/* Each chunk has an identifier and a length.  The length gives the number of
   bytes in the chunk (not including the ckID or ckDataSize fields).  If the
   ckDataSize field is odd, a zero byte fills the space before the start of the
   next chunk.
   - Some AIFF-C files do not have this zero byte at the end of the file,
     i.e. the total file length is odd, yet the FORM chunk size indicates that
     the padding byte is there.
   - The chunks after the FORM chunk can appear in any order.
   - The standard is written in such a way that it is implied that one should
     scan the whole file for chunks.  Here we are trying to allow for non-
     seekable input.
     - If the COMM chunk comes ahead of the SSND chunk all is fine.  We get to
       the SSND chunk and then skip forward to the data.
     - If the COMM chunk follows the SSND chunk we will have to seek back to
       the data in the SSND chunk.  The file has to be seekable.
     - If the number of samples is zero (in the COMM chunk), there need be no
       SSND chunk.  When we get to the end-of-file, we will declare this to be
       the start of data.
*/

/* Defaults and inital values */
  AFr.DFormat.Format = FD_UNDEF;
  AFr.NData.Ldata = AF_LDATA_UNDEF;
  AFr.Hinfo.Info = Info;
  AFr.Hinfo.N = 0;
  AFr.Hinfo.Nmax = AF_MAXINFO;

/* Check the FORM chunk */
  LFORM = AF_rdFORM_AIFF (fp, &Ftype);
  offs = 12L;	/* Positioned after FORM/AIFF preamble */
  
  Dstart = 0L;
  AtData = 0;
  while (offs < LFORM) {

    /* Read the chunk preamble */
    offs += RHEAD_S (fp, CkHead.ckID);
    offs += RHEAD_V (fp, CkHead.ckDataSize, DS_EB);

    /* COMM chunk */
    if (SAME_CSTR (CkHead.ckID, CKID_COMM)) {
      offs += AF_rdCOMM (fp, (int) CkHead.ckDataSize, Ftype, &CkCOMM);
      AF_decCOMM (&CkCOMM, &AFr);
    }

    /* FVER chunk */
    else if (Ftype == FT_AIFF_C && SAME_CSTR (CkHead.ckID, CKID_FVER))
      offs += AF_checkFVER (fp, (int) CkHead.ckDataSize);

    /* Text chunks */
    else if (SAME_CSTR (CkHead.ckID, CKID_NAME))
      offs += AFrdHtext (fp, (int) CkHead.ckDataSize, "name: ",
			 &AFr.Hinfo, ALIGN);
    else if (SAME_CSTR (CkHead.ckID, CKID_AUTHOR))
      offs += AFrdHtext (fp, (int) CkHead.ckDataSize, "author: ",
			 &AFr.Hinfo, ALIGN);
    else if (SAME_CSTR (CkHead.ckID, CKID_COPYRIGHT))
      offs += AFrdHtext (fp, (int) CkHead.ckDataSize, "copyright: ",
			 &AFr.Hinfo, ALIGN);

    /* Annotation chunk */
    else if (SAME_CSTR (CkHead.ckID, CKID_ANNOTATION)) {
      N = AFr.Hinfo.N;
      offs += AFrdHinfo (fp, (int) CkHead.ckDataSize, &AFr.Hinfo, ALIGN);
      if (AFr.Hinfo.N == N)
	offs += AFrdHtext (fp, (int) CkHead.ckDataSize, "annotation: ",
			   &AFr.Hinfo, ALIGN);
    }

    /* SSND chunk */
    else if (SAME_CSTR (CkHead.ckID, CKID_SSND)) {
      offs += AF_decSSND (fp, (long int) CkHead.ckDataSize, &AFr.NData.Ldata);
      Dstart = offs;
      if (RNDUPV (AFr.NData.Ldata + offs, ALIGN) >= LFORM ||
	  ! FLseekable (fp)) {
	AtData = 1;
	break;
      }
      else {
	AtData = 0;
	offs += RSKIP (fp, RNDUPV (offs + AFr.NData.Ldata, ALIGN) - offs);
      }
    }

    /* Miscellaneous chunks */
    else
      offs += RSKIP (fp, RNDUPV (CkHead.ckDataSize, ALIGN));
 
  }
  if (AFr.DFormat.Format == FD_UNDEF) {
    UTwarn ("AFrdAIhead - %s", AFM_AIFF_BadHead);
    return NULL;
  }
  if (! AtData && offs != LFORM)
    UTwarn ("AFrdAIhead - %s", AFM_AIFF_FixFORM);

  /* If no SSND chunk has been found, check for a zero data size (does not
     need a SSND chunk)
  */
  if (Dstart == 0L) {
    if (CkCOMM.numSampleFrames == 0) {
      AFr.NData.Ldata = 0L;
      AtData = 1;
      Dstart = offs;
    }
    else {
      UTwarn ("AFrdAIhead - %s", AFM_AIFF_NoSSND);
      return NULL;
    }
  }

  /* Position at the start of data */
  if (! AtData) {
    if (AFseek (fp, Dstart, NULL))
      return NULL;
  }

/* Set the parameters for file access */
  AFp = AFsetRead (fp, Ftype, &AFr, AF_NOFIX);

  return AFp;
}

/* Check the RIFF/AIFF file preamble:  On error return via longjmp */


static long int
AF_rdFORM_AIFF (FILE *fp, int *Ftype)

{
  int offs;
  long int LFORM, Lfile;
  struct AI_CkPreamb CkHead;

  offs = RHEAD_S (fp, CkHead.ckID);
  if (! SAME_CSTR (CkHead.ckID, FM_IFF)) {
    UTwarn ("AFrdAIhead - %s", AFM_AIFF_BadId);
    longjmp (AFR_JMPENV, 1);
  }

  offs += RHEAD_V (fp, CkHead.ckDataSize, DS_EB);
  LFORM = CkHead.ckDataSize + offs;

  if (FLseekable (fp)) {
    Lfile = FLfileSize (fp);
    if (LFORM > Lfile) {
      LFORM = Lfile;
      UTwarn ("AFrdAIhead - %s", AFM_AIFF_FixFORM);
    }
  }
 
  offs += RHEAD_S (fp, CkHead.ckID);
  if (SAME_CSTR (CkHead.ckID, FM_AIFF_C))
    *Ftype = FT_AIFF_C;
  else if (SAME_CSTR (CkHead.ckID, FM_AIFF))
    *Ftype = FT_AIFF;
  else {
    UTwarn ("AFrdAIhead - %s", AFM_AIFF_BadId);
    longjmp (AFR_JMPENV, 1);
  }

  return LFORM;
}

/* Read the COMM chunk */


static int
AF_rdCOMM (FILE *fp, int Size, int Ftype, struct AI_CkCOMM *CkCOMM)

{
  int offs;

  offs = RHEAD_V (fp, CkCOMM->numChannels, DS_EB);
  offs += RHEAD_V (fp, CkCOMM->numSampleFrames, DS_EB);
  offs += RHEAD_V (fp, CkCOMM->sampleSize, DS_EB);
  offs += RHEAD_S (fp, CkCOMM->sampleRate);

  /* Set the compressionType field for both AIFF and AIFF-C files */
  if (Ftype == FT_AIFF_C) {
    offs += RHEAD_S (fp, CkCOMM->compressionType);
    offs += AF_rPstring (fp, CkCOMM->compressionName, CNAME_MAX);
  }
  else {
    memcpy (CkCOMM->compressionType, CT_NONE, sizeof (CT_NONE));
    CkCOMM->compressionName[0] = '\0';
  }

  /* Some AIFF-C COMM chunks are declared to be of length 18 (the size for
     AIFF files) when they are in fact larger.  Here we find the actual
     length of the COMM chunk.
  */
  if (offs > Size) {
    UTwarn ("AFrdAIhead - %s", AFM_AIFF_FixCOMM);
    Size = offs;
  }
  offs += RSKIP (fp, RNDUPV (Size, ALIGN) - offs);

  return offs;
}

/* Read a P-string */


static int
AF_rPstring (FILE *fp, char string[], int ncMax)

{
  int offs, nc, ncP, nr;
  char slen[1];
 
  offs = RHEAD_S (fp, slen);	/* 1 byte length */

  nc = (int) slen[0];
  ncP = RNDUPV (nc + 1, 2);
  nr = MINV (nc, ncMax);

  offs += RHEAD_SN (fp, string, nr);
  string[nr] = '\0';
  offs += RSKIP (fp, ncP - (nc + 1));

  return offs;
}

/* Decode the data format: On error return via longjmp */


static void
AF_decCOMM (struct AI_CkCOMM *CkCOMM, struct AF_read *AFr)

{
  /* Uncompressed */
  if (SAME_CSTR (CkCOMM->compressionType, CT_NONE)) {
    if (CkCOMM->sampleSize > 0 && CkCOMM->sampleSize <= 8) {
      AFr->DFormat.Format = FD_INT8;
      AFr->DFormat.ScaleF = AI_SF_NONE8;
    }
    else if (CkCOMM->sampleSize > 8 && CkCOMM->sampleSize <= 16) {
      AFr->DFormat.Format = FD_INT16;
      AFr->DFormat.ScaleF = AI_SF_NONE16;
    }
    else if (CkCOMM->sampleSize > 16 && CkCOMM->sampleSize <= 24) {
      AFr->DFormat.Format = FD_INT24;
      AFr->DFormat.ScaleF = AI_SF_NONE24;
    }
    else if (CkCOMM->sampleSize > 24 && CkCOMM->sampleSize <= 32) {
      AFr->DFormat.Format = FD_INT32;
      AFr->DFormat.ScaleF = AI_SF_NONE32;
    }
    else {
      UTwarn ("AFrdAIhead - %s: \"%d\"", AFM_AIFF_UnsSSize,
	      (int) CkCOMM->sampleSize);
      longjmp (AFR_JMPENV, 1);
    }
  }

  /* AIFF-C, compressed */
  /* A-law and mu-law are compressed formats; for these formats
   CkCOMM.sampleSize = 16 (not checked) */
  else if (SAME_CSTR (CkCOMM->compressionType, CT_ULAW) ||
	   SAME_CSTR (CkCOMM->compressionType, CT_X_ULAW)) {
    AFr->DFormat.Format = FD_MULAW8;
    AFr->DFormat.ScaleF = AI_SF_ULAW;
  }
  else if (SAME_CSTR (CkCOMM->compressionType, CT_ALAW) ||
	   SAME_CSTR (CkCOMM->compressionType, CT_X_ALAW)) {
    AFr->DFormat.Format = FD_ALAW8;
    AFr->DFormat.ScaleF = AI_SF_ALAW;
  }
  else if (SAME_CSTR (CkCOMM->compressionType, CT_FLOAT32) ||
	   SAME_CSTR (CkCOMM->compressionType, CT_FL32)) {
    AFr->DFormat.Format = FD_FLOAT32;
    AFr->DFormat.ScaleF = AI_SF_FLOAT32;
    if (! UTcheckIEEE ())
      UTwarn ("AFrdAIhead - %s", AFM_AIFF_NoIEEE);
  }
  else if (SAME_CSTR (CkCOMM->compressionType, CT_FLOAT64)) {
    AFr->DFormat.Format = FD_FLOAT64;
    AFr->DFormat.ScaleF = AI_SF_FLOAT64;
    if (! UTcheckIEEE ())
      UTwarn ("AFrdAIhead - %s", AFM_AIFF_NoIEEE);
  }
  else {
    UTwarn ("AFrdAIhead - %s: \"%.4s\" (\"%s\")", AFM_AIFF_UnsComp,
	    CkCOMM->compressionType, STstrDots (CkCOMM->compressionName, 32));
    longjmp (AFR_JMPENV, 1);
  }
  AFr->DFormat.Swapb = DS_EB;

  AFr->Sfreq = UTdIEEE80 (CkCOMM->sampleRate);
  AFr->NData.Nsamp = CkCOMM->numSampleFrames * CkCOMM->numChannels;
  AFr->NData.Nchan = CkCOMM->numChannels;

  return;
} 

/* Check the format version */


static int
AF_checkFVER (FILE *fp, int Size)

{
  int offs;
  struct AI_CkFVER CkFVER;
  
  offs = RHEAD_V (fp, CkFVER.timestamp, DS_EB);
  if (CkFVER.timestamp != AIFCVersion1)
    UTwarn ("AFrdAIhead - %s", AFM_AIFF_BadVer);
  offs += RSKIP (fp, RNDUPV (Size, ALIGN) - offs);

  return offs;
}

/* Decode the SSND chunk
   Position left at the beginning of data or at the end of the SSND chunk
*/


static long int
AF_decSSND (FILE *fp, long int Size, long int *Ldata)

{
  long int offs;
  struct AI_CkSSND CkSSND;

  /* Data size, skip to start of data */
  offs  = RHEAD_V (fp, CkSSND.offset, DS_EB);
  offs += RHEAD_V (fp, CkSSND.blockSize, DS_EB);
  offs += RSKIP (fp, CkSSND.offset);

  *Ldata = Size - offs;

  return offs;
}
